# Copyright 2019 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import struct
import base64

# constants
ITEM_HEADER_MAGIC = struct.pack("I", 11223344)
ITEM_HEADER_MAGIC_AND_VERSION = ITEM_HEADER_MAGIC + struct.pack("I", 1)

OPERAND_TYPE_LONG = 259
OPERAND_TYPE_DOUBLE = 261


def encode_list(list_value):
    typecode = 'l'
    if len(list_value) and isinstance(list_value[0], float):
        typecode = 'd'

    return encode_array(list_value, typecode)


def encode_array(array_value, typecode):
    num_items = len(array_value)
    operand_type = OPERAND_TYPE_LONG if typecode == 'l' else OPERAND_TYPE_DOUBLE

    encoded_array = ITEM_HEADER_MAGIC_AND_VERSION + struct.pack('II' + typecode * num_items,
                                                                num_items * 8,
                                                                operand_type,
                                                                *array_value)

    return base64.b64encode(encoded_array)


def decode(encoded_array):
    static_header_len = len(ITEM_HEADER_MAGIC_AND_VERSION)

    # do a quick peek before we decode
    if len(encoded_array) <= static_header_len or not encoded_array.startswith(ITEM_HEADER_MAGIC_AND_VERSION):
        raise ValueError('Not an encoded array')

    # get header (which contains number of items and type
    header = encoded_array[static_header_len:static_header_len + 8]
    values = encoded_array[static_header_len + len(header):]

    # unpack the header to get the size and operand
    unpacked_header = struct.unpack('II', header)

    # get the typecode and number of items
    typecode = 'l' if unpacked_header[1] == OPERAND_TYPE_LONG else 'd'
    num_items = int(unpacked_header[0] / 8)

    # decode the values
    return list(struct.unpack(typecode * num_items, values))



# Copyright 2019 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import base64
import future.utils
import os
import array
import datetime

try:
    from urllib.parse import urlencode, quote
except BaseException:
    from urllib import urlencode, quote

import ujson

import v3io.common.helpers
import v3io.dataplane.kv_array
import v3io.dataplane.kv_timestamp


#
# Request
#

class Request(object):

    def __init__(self,
                 container,
                 access_key,
                 raise_for_status,
                 encoder,
                 encoder_args,
                 output=None):
        self.container = container
        self.access_key = access_key
        self.raise_for_status = raise_for_status
        self.encoder = encoder
        self.encoder_args = encoder_args
        self.output = output

        # get request params with the encoder
        self.method, self.path, self.query, self.headers, self.body = encoder(container, access_key, encoder_args)

        # used by the transport
        self.transport = lambda: None

    def encode_path(self):
        if self.query is None:
            return quote(self.path)

        return quote(self.path) + '?' + urlencode(self.query, quote_via=quote)


#
# Encoders
#


#
# Container
#

def encode_get_containers(container_name, access_key, kwargs):
    return _encode('GET', '/', access_key, None, None, {}, None)


def encode_get_container_contents(container_name, access_key, kwargs):
    query = {
        'prefix': kwargs['path']
    }

    if kwargs['get_all_attributes']:
        query['prefix-info'] = 1

    if kwargs['directories_only']:
        query['prefix-only'] = 1

    if kwargs['limit'] is not None:
        query['max-keys'] = kwargs['limit']

    if kwargs['marker']:
        query['marker'] = kwargs['marker']

    return _encode('GET',
                   '/' + container_name,
                   access_key,
                   None,
                   query,
                   {},
                   None)


#
# Object
#

def encode_head_object(container_name, access_key, kwargs):
    return _encode('HEAD', container_name, access_key, kwargs['path'], None, None, None)


def encode_get_object(container_name, access_key, kwargs):
    headers = None

    offset = kwargs.get('offset')

    # if the append flag is passed, add a range header
    if offset:
        range_value = 'bytes=' + str(offset)

        num_bytes = kwargs.get('num_bytes')
        if num_bytes:
            range_value += '-' + str(offset + num_bytes - 1)

        headers = {
            'Range': range_value
        }

    return _encode('GET', container_name, access_key, kwargs['path'], None, headers, None)


def encode_put_object(container_name, access_key, kwargs):
    headers = None

    # if the append flag is passed, add a range header
    if kwargs['append']:
        headers = {
            'Range': '-1'
        }

    return _encode('PUT', container_name, access_key, kwargs['path'], None, headers, kwargs['body'])


def encode_delete_object(container_name, access_key, kwargs):
    return _encode('DELETE', container_name, access_key, kwargs['path'], None, None, None)


#
# KV
#

def encode_put_item(container_name, access_key, kwargs):
    # add 'Item' to body
    body = {
        'Item': _dict_to_typed_attributes(kwargs['attributes'])
    }

    if kwargs['condition'] is not None:
        body['ConditionExpression'] = kwargs['condition']

    return _encode('PUT',
                   container_name,
                   access_key,
                   kwargs.get('path') or os.path.join(kwargs['table_path'], kwargs['key']),
                   None,
                   {'X-v3io-function': 'PutItem'},
                   body)


def encode_update_item(container_name, access_key, kwargs):
    body = {
        'UpdateMode': kwargs.get('update_mode') or 'CreateOrReplaceAttributes'
    }

    if kwargs['condition'] is not None:
        body['ConditionExpression'] = kwargs['condition']

    if not kwargs['expression'] and not kwargs['attributes']:
        raise RuntimeError('One of expression or attributes must be populated for update item')

    if kwargs['expression']:
        http_method = 'POST'
        function_name = 'UpdateItem'
        body['UpdateExpression'] = kwargs['expression']

    if kwargs['alternate_expression']:
        http_method = 'POST'
        function_name = 'UpdateItem'
        body['AlternateUpdateExpression'] = kwargs['alternate_expression']

    elif kwargs['attributes']:
        http_method = 'PUT'
        function_name = 'PutItem'
        body['Item'] = _dict_to_typed_attributes(kwargs['attributes'])

    return _encode(http_method,
                   container_name,
                   access_key,
                   kwargs.get('path') or os.path.join(kwargs['table_path'], kwargs['key']),
                   None,
                   {'X-v3io-function': function_name},
                   body)


def encode_get_item(container_name, access_key, kwargs):
    body = {
        'AttributesToGet': ','.join(kwargs['attribute_names'])
    }

    return _encode('PUT',
                   container_name,
                   access_key,
                   kwargs.get('path') or os.path.join(kwargs['table_path'], kwargs['key']),
                   None,
                   {'X-v3io-function': 'GetItem'},
                   body)


def encode_get_items(container_name, access_key, kwargs):
    body = {
        'AttributesToGet': ','.join(kwargs['attribute_names']),
    }

    if kwargs.get('table_name'):
        body['TableName'] = kwargs['table_name']

    if kwargs['filter_expression']:
        body['FilterExpression'] = kwargs['filter_expression']

    if kwargs['marker']:
        body['Marker'] = kwargs['marker']

    if kwargs['sharding_key']:
        body['ShardingKey'] = kwargs['sharding_key']

    if kwargs['limit'] is not None:
        body['Limit'] = kwargs['limit']

    if kwargs['segment'] is not None:
        body['Segment'] = kwargs['segment']

    if kwargs['total_segments'] is not None:
        body['TotalSegment'] = kwargs['total_segments']

    if kwargs['sort_key_range_start']:
        body['SortKeyRangeStart'] = kwargs['sort_key_range_start']

    if kwargs['sort_key_range_end']:
        body['SortKeyRangeEnd'] = kwargs['sort_key_range_end']

    return _encode('PUT',
                   container_name,
                   access_key,
                   kwargs.get('path') or kwargs['table_path'],
                   None,
                   {'X-v3io-function': 'GetItems'},
                   body)


#
# Stream
#

def encode_create_stream(container_name, access_key, kwargs):
    body = {
        'ShardCount': kwargs['shard_count'],
        'RetentionPeriodHours': kwargs.get('retention_period_hours') or 24
    }

    return _encode('POST',
                   container_name,
                   access_key,
                   kwargs.get('path') or kwargs['stream_path'],
                   None,
                   {'X-v3io-function': 'CreateStream'},
                   body)


def encode_update_stream(container_name, access_key, kwargs):
    body = {
        'ShardCount': kwargs['shard_count'],
    }

    return _encode('POST',
                   container_name,
                   access_key,
                   kwargs.get('path') or kwargs['stream_path'],
                   None,
                   {'X-v3io-function': 'UpdateStream'},
                   body)


def encode_describe_stream(container_name, access_key, kwargs):
    return _encode('PUT',
                   container_name,
                   access_key,
                   kwargs.get('path') or kwargs['stream_path'],
                   None,
                   {'X-v3io-function': 'DescribeStream'},
                   None)


def encode_seek_shard(container_name, access_key, kwargs):
    body = {
        'Type': kwargs['seek_type'],
    }

    if kwargs['seek_type'] == 'SEQUENCE':
        body['StartingSequenceNumber'] = kwargs['starting_sequence_number']
    elif kwargs['seek_type'] == 'TIME':
        body['TimestampSec'] = kwargs['timestamp_sec']
        body['TimestampNSec'] = kwargs['timestamp_nsec']
    elif kwargs['seek_type'] not in ['EARLIEST', 'LATEST']:
        raise ValueError('Unsupported seek_type ({0}) for seek_shard. Must be one of SEQUENCE, TIME, EARLIEST, LATEST'.
                         format(kwargs['seek_type']))

    return _encode('PUT',
                   container_name,
                   access_key,
                   kwargs.get('path') or kwargs['stream_path'],
                   None,
                   {'X-v3io-function': 'SeekShard'},
                   body)


def encode_put_records(container_name, access_key, kwargs):
    records = []

    for record in kwargs['records']:
        record_body = {
            'Data': _to_base64(record['data']),
        }

        try:
            record_body['ClientInfo'] = _to_base64(record['client_info'])
        except KeyError:
            pass

        try:
            record_body['ShardId'] = record['shard_id']
        except KeyError:
            pass

        try:
            record_body['PartitionKey'] = record['partition_key']
        except KeyError:
            pass

        records.append(record_body)

    body = {
        'Records': records
    }

    return _encode('POST',
                   container_name,
                   access_key,
                   kwargs.get('path') or kwargs['stream_path'],
                   None,
                   {'X-v3io-function': 'PutRecords'},
                   body)


def encode_get_records(container_name, access_key, kwargs):
    body = {
        'Location': kwargs['location'],
    }

    limit = kwargs.get('limit')
    if limit:
        body['Limit'] = limit

    return _encode('PUT',
                   container_name,
                   access_key,
                   kwargs.get('path') or kwargs['stream_path'],
                   None,
                   {'X-v3io-function': 'GetRecords'},
                   body)


#
# Helpers
#

def _encode(method, container_name, access_key, path, query, headers, body):
    if path is not None:
        path = v3io.common.helpers.url_join(container_name, path)
    else:
        path = container_name

    headers, body = _resolve_body_and_headers(access_key, headers, body)

    return method, path, query, headers, body


def _typed_attributes_to_dict(self):
    pass


def _ensure_trailing_slash(path):
    if not path.endswith('/'):
        return path + '/'

    return path


def _to_base64(input):
    if isinstance(input, str):
        input = input.encode('utf-8')

    return base64.b64encode(input)


def _dict_to_typed_attributes(d):
    typed_attributes = {}

    for (key, value) in future.utils.viewitems(d):
        attribute_type = type(value)
        type_value = None

        if isinstance(value, future.utils.text_type):
            type_key = 'S'
            type_value = value
        elif isinstance(value, future.utils.string_types):
            type_key = 'S'
            type_value = str(value)
        elif attribute_type in [int, float]:
            type_key = 'N'
            type_value = str(value)
        elif attribute_type in [bytes, bytearray]:
            type_key = 'B'
            type_value = base64.b64encode(value)
        elif isinstance(value, bool):
            type_key = 'BOOL'
            type_value = value
        elif isinstance(value, list):
            type_key = 'B'
            type_value = v3io.dataplane.kv_array.encode_list(value)
        elif isinstance(value, array.array):
            type_key = 'B'
            type_value = v3io.dataplane.kv_array.encode_array(value, value.typecode)
        elif isinstance(value, datetime.datetime):
            type_key = 'TS'
            type_value = v3io.dataplane.kv_timestamp.encode(value)
        else:
            raise AttributeError('Attribute {0} has unsupported type {1}'.format(key, attribute_type))

        typed_attributes[key] = {type_key: type_value}

    return typed_attributes


def _resolve_body_and_headers(access_key, headers, body):
    if access_key:
        headers = headers or {}
        headers['X-v3io-session-key'] = access_key

    if not isinstance(body, dict):
        return headers, body

    body = ujson.dumps(body, reject_bytes=False)
    headers['Content-Type'] = 'application/json'

    return headers, body



# Copyright 2019 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import ssl
import sys
import http.client
import socket

import v3io.dataplane.response
import v3io.dataplane.request
from . import abstract
import queue


class Transport(abstract.Transport):

    def __init__(self, logger, endpoint=None, max_connections=None, timeout=None, verbosity=None):
        super(Transport, self).__init__(logger, endpoint, max_connections, timeout, verbosity)

        self._free_connections = queue.Queue()

        # based on scheme, create a host and context for _create_connection
        self._host, self._ssl_context = self._parse_endpoint(self._endpoint)

        # create the pool connection
        self._create_connections(self.max_connections,
                                 self._host,
                                 self._ssl_context)

        # python 2 and 3 have different exceptions
        if sys.version_info[0] >= 3:
            self._wait_response_exceptions = (
                http.client.RemoteDisconnected, ConnectionResetError, ConnectionRefusedError, http.client.ResponseNotReady)
            self._send_request_exceptions = (
                BrokenPipeError, http.client.CannotSendRequest, http.client.RemoteDisconnected)
            self._get_status_and_headers = self._get_status_and_headers_py3
        else:
            self._wait_response_exceptions = (http.client.BadStatusLine, socket.error)
            self._send_request_exceptions = (http.client.CannotSendRequest, http.client.BadStatusLine)
            self._get_status_and_headers = self._get_status_and_headers_py2

    def requires_access_key(self):
        return True

    def send_request(self, request):
        # TODO: consider getting param of whether we should block or not (wait for connection to be free or raise exception)
        connection = self._free_connections.get(block=True, timeout=None)

        try:
            return self._send_request_on_connection(request, connection)
        except BaseException as e:
            connection.close()
            connection = self._create_connection(self._host, self._ssl_context)
            self._free_connections.put(connection, block=True)
            raise e

    def wait_response(self, request, raise_for_status=None, num_retries=1):
        connection = request.transport.connection_used

        while True:
            try:

                # read the response
                response = connection.getresponse()
                response_body = response.read()

                status_code, headers = self._get_status_and_headers(response)

                self.log('Rx',
                         connection=connection,
                         status_code=status_code,
                         body=response_body)

                response = v3io.dataplane.response.Response(request.output,
                                                            status_code,
                                                            headers,
                                                            response_body)

                # enforce raise for status
                response.raise_for_status(request.raise_for_status or raise_for_status)

                # return the response
                return response

            except v3io.dataplane.response.HttpResponseError as response_error:
                self._logger.warn_with('Response error: {}'.format(str(response_error)))
                raise response_error
            except BaseException as e:
                if num_retries == 0:
                    self._logger.error_with('Error occurred while waiting for response and ran out of retries',
                                            e=type(e),
                                            e_msg=e,
                                            response_body=response_body,
                                            status_code=status_code,
                                            headers=headers,
                                            connection=connection)

                    raise e

                self._logger.debug_with('Error occurred while waiting for response – retrying',
                                        retries_left=num_retries,
                                        e=type(e),
                                        e_msg=e,
                                        connection=connection)

                num_retries -= 1

                connection.close()
                connection = self._create_connection(self._host, self._ssl_context)

                # re-send the request on the connection
                request = self._send_request_on_connection(request, connection)
            finally:
                self._free_connections.put(connection, block=True)

    def _send_request_on_connection(self, request, connection):
        setattr(request.transport, 'connection_used', connection)

        path = request.encode_path()

        self.log('Tx',
                 connection=connection,
                 method=request.method,
                 path=path,
                 headers=request.headers,
                 body=request.body)

        try:
            try:
                connection.request(request.method, path, request.body, request.headers)
            except self._send_request_exceptions as e:
                self._logger.debug_with('Disconnected while attempting to send. Recreating connection and retrying',
                                        e=type(e), e_msg=e, connection=connection)
                connection.close()
                connection = self._create_connection(self._host, self._ssl_context)
                request.transport.connection_used = connection
                connection.request(request.method, path, request.body, request.headers)
        except BaseException as e:
            self._logger.error_with('Unhandled exception while sending request', e=type(e), e_msg=e, connection=connection)
            raise e

        return request

    def _create_connections(self, num_connections, host, ssl_context):
        for _ in range(num_connections):
            connection = self._create_connection(host, ssl_context)
            self._free_connections.put(connection, block=True)

    def _create_connection(self, host, ssl_context):
        if ssl_context is None:
            return http.client.HTTPConnection(host)

        return http.client.HTTPSConnection(host, context=ssl_context)

    def _parse_endpoint(self, endpoint):
        if endpoint.startswith('http://'):
            return endpoint[len('http://'):], None

        if endpoint.startswith('https://'):
            ssl_context = ssl.create_default_context()
            ssl_context.check_hostname = False
            ssl_context.verify_mode = ssl.CERT_NONE

            return endpoint[len('https://'):], ssl_context

        return endpoint, None

    def _get_status_and_headers_py2(self, response):
        return response.status, response.getheaders()

    def _get_status_and_headers_py3(self, response):
        return response.code, response.headers



# Copyright 2019 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from . import abstract


class Transport(abstract.Transport):

    def __init__(self, request_verifiers):
        super().__init__(None, '', 0, None, 'DEBUG')
        self._request_verifiers = request_verifiers
        self._current_request_index = 0

    def close(self):
        pass

    def wait_response(self, request, raise_for_status=None):
        if self._current_request_index > len(self._request_verifiers):
            raise IndexError(f'Have only {len(self._request_verifiers)} verifiers, got {self._current_request_index} requests')

        # call the verifier, get the response
        response = self._request_verifiers[self._current_request_index](request)

        self._current_request_index += 1

        return response



# Copyright 2019 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import v3io.dataplane.request
import v3io.dataplane.output
import v3io.dataplane.model
import v3io.dataplane.kv_cursor


class Model(v3io.dataplane.model.Model):

    def __init__(self, client):
        self._client = client
        self._access_key = client._access_key
        self._transport = client._transport

    async def list(self,
                   container,
                   path,
                   access_key=None,
                   raise_for_status=None,
                   get_all_attributes=None,
                   directories_only=None,
                   limit=None,
                   marker=None):
        """Lists the containers contents.

        Parameters
        ----------
        container (Required) : str
            The container on which to operate.
        path (Required) : str
            The path within the container
        access_key (Optional) : str
            The access key with which to authenticate. Defaults to the V3IO_ACCESS_KEY env.
        get_all_attributes (Optional) : bool
            False (default) - retrieves basic attributes
            True - retrieves all attributes of the underlying objects
        directories_only (Optional) : bool
            False (default) - retrieves objects (contents) and directories (common prefixes)
            True - retrieves only directories (common prefixes)
        limit (Optional) : int
            Number of objects/directories to receive. default: 1000
        marker (Optional) : str
            An opaque identifier that was returned in the NextMarker element of a response to a previous
            get_container_contents request that did not return all the requested items. This marker identifies the
            location in the path from which to start searching for the remaining requested items.

        Return Value
        ----------
        A `Response` object, whose `output` is `GetContainerContentsOutput`.
        """
        return await self._transport.request(container,
                                             access_key or self._access_key,
                                             raise_for_status,
                                             v3io.dataplane.request.encode_get_container_contents,
                                             locals(),
                                             v3io.dataplane.output.GetContainerContentsOutput)



# Copyright 2019 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os

import v3io.dataplane.request
import v3io.dataplane.output
import v3io.dataplane.model
import v3io.dataplane.kv_cursor


class Model(v3io.dataplane.model.Model):

    def __init__(self, client):
        self._client = client
        self._access_key = client._access_key
        self._transport = client._transport

    def create(self,
               container,
               stream_path,
               shard_count,
               access_key=None,
               raise_for_status=None,
               transport_actions=None,
               retention_period_hours=None):
        """Creates and configures a new stream. The configuration includes the stream's shard count and retention
        period. The new stream is available immediately upon its creation.

        See https://www.iguazio.com/docs/latest-release/data-layer/reference/web-apis/streaming-web-api/createstream/.

        Parameters
        ----------
        container (Required) : str
            The container on which to operate.
        stream_path (Required) : str
            A unique name for the new stream (collection) that will be created.
        shard_count (Required) : int
            The steam's shard count, i.e., the number of stream shards to create.
        access_key (Optional) : str
            The access key with which to authenticate. Defaults to the V3IO_ACCESS_KEY env.
        retention_period_hours (Optional) : int
            The stream's retention period, in hours. After this period elapses, when new records are added to the
            stream, the earliest ingested records are deleted. default: 24

        Return Value
        ----------
        A `Response` object.
        """
        stream_path = self._ensure_path_ends_with_slash(stream_path)

        return self._transport.request(container,
                                       access_key or self._access_key,
                                       raise_for_status,
                                       transport_actions,
                                       v3io.dataplane.request.encode_create_stream,
                                       locals())

    def update(self,
               container,
               stream_path,
               shard_count,
               access_key=None,
               raise_for_status=None,
               transport_actions=None):
        """Updates a stream's configuration by increasing its shard count. The changes are applied immediately.

        See https://www.iguazio.com/docs/latest-release/data-layer/reference/web-apis/streaming-web-api/updatestream/.

        Parameters
        ----------
        container (Required) : str
            The container on which to operate.
        stream_path (Required) : str
            A unique name for the new stream (collection) that will be created.
        shard_count (Required) : int
            The steam's shard count, i.e., the number of stream shards to create.
        access_key (Optional) : str
            The access key with which to authenticate. Defaults to the V3IO_ACCESS_KEY env.

        Return Value
        ----------
        A `Response` object.
        """
        stream_path = self._ensure_path_ends_with_slash(stream_path)

        return self._transport.request(container,
                                       access_key or self._access_key,
                                       raise_for_status,
                                       transport_actions,
                                       v3io.dataplane.request.encode_update_stream,
                                       locals())

    def delete(self, container, stream_path, access_key=None, raise_for_status=None):
        """Deletes a stream object along with all of its shards.

        Parameters
        ----------
        container (Required) : str
            The container on which to operate.
        stream_path (Required) : str
            The stream_path of the stream.
        access_key (Optional) : str
            The access key with which to authenticate. Defaults to the V3IO_ACCESS_KEY env.

        Return Value
        ----------
        A `Response` object.
        """
        stream_path = self._ensure_path_ends_with_slash(stream_path)

        response = self._client.container.list(container,
                                               stream_path,
                                               access_key,
                                               raise_for_status)

        # nothing to do
        if response.status_code == 404:
            return response

        for stream_shard in response.output.contents:
            self._client.object.delete(container, stream_shard.key, access_key, raise_for_status)

        return self._client.object.delete(container, stream_path, access_key, raise_for_status)

    def describe(self, container, stream_path, access_key=None, raise_for_status=None, transport_actions=None):
        """Retrieves a stream's configuration, including the shard count and retention period.

        See https://www.iguazio.com/docs/latest-release/data-layer/reference/web-apis/streaming-web-api/describestream/.

        Parameters
        ----------
        stream_path (Required) : str
            The stream_path of the stream.
        access_key (Optional) : str
            The access key with which to authenticate. Defaults to the V3IO_ACCESS_KEY env.

        Return Value
        ----------
        A `Response` object, whose `output` is `DescribeStreamOutput`.
        """
        stream_path = self._ensure_path_ends_with_slash(stream_path)

        return self._transport.request(container,
                                       access_key or self._access_key,
                                       raise_for_status,
                                       transport_actions,
                                       v3io.dataplane.request.encode_describe_stream,
                                       locals(),
                                       v3io.dataplane.output.DescribeStreamOutput)

    def seek(self,
             container,
             stream_path,
             shard_id,
             seek_type,
             access_key=None,
             raise_for_status=None,
             transport_actions=None,
             starting_sequence_number=None,
             timestamp_sec=None,
             timestamp_nsec=None):
        """Returns the requested location within the specified stream shard, for use in a subsequent GetRecords
        operation. The operation supports different seek types, as outlined in the Stream Record Consumption
        overview and in the description of the Type request parameter below.

        See https://www.iguazio.com/docs/latest-release/data-layer/reference/web-apis/streaming-web-api/seek/.

        Parameters
        ----------
        container (Required) : str
            The container on which to operate.
        stream_path (Required) : str
            The stream_path of the stream.
        seek_type (Required) : str
            'EARLIEST': the location of the earliest ingested record in the shard.
            'LATEST': the location of the end of the shard.
            'TIME': the location of the earliest ingested record in the shard beginning at the base time set in the
                    TimestampSec and TimestampNSec request parameters. If no matching record is found (i.e., if all
                    records in the shard arrived before the specified base time) the operation returns the location
                    of the end of the shard.
            'SEQUENCE': the location of the record whose sequence number matches the sequence number specified in the
                        StartingSequenceNumber request parameter. If no match is found, the operation fails.
        access_key (Optional) : str
            The access key with which to authenticate. Defaults to the V3IO_ACCESS_KEY env.
        starting_sequence_number (Optional) : int
            Record sequence number for a sequence-number based seek operation - Type=SEQUENCE. When this parameter is
            set, the operation returns the location of the record whose sequence number matches the parameter.
        timestamp_sec (Optional) : int
            The base time for a time-based seek operation (Type=TIME), as a Unix timestamp in seconds. For example,
            1511260205 sets the search base time to 21 Nov 2017 at 10:30:05 AM UTC. The TimestampNSec request parameter
            sets the nanoseconds unit of the seek base time.

            When the TimestampSec and TimestampNSec parameters are set, the operation searches for the location of the
            earliest ingested record in the shard (the earliest record that arrived at the platform) beginning at the
            specified base time. If no matching record is found (i.e., if all records in the shard arrived before the
            specified base time), return the last location in the shard.
        timestamp_nsec (Optional) : int
            The nanoseconds unit of the TimestampSec base-time timestamp for a time-based seek operation (Type=TIME).
            For example, if TimestampSec is 1511260205 and TimestampNSec is 500000000, seek should search for the
            earliest ingested record since 21 Nov 2017 at 10:30 AM and 5.5 seconds.

        Return Value
        ----------
        A `Response` object, whose `output` is `SeekShardOutput`.
        """
        stream_path = self._ensure_path_ends_with_slash(os.path.join(stream_path, str(shard_id)))

        return self._transport.request(container,
                                       access_key or self._access_key,
                                       raise_for_status,
                                       transport_actions,
                                       v3io.dataplane.request.encode_seek_shard,
                                       locals(),
                                       v3io.dataplane.output.SeekShardOutput)

    def put_records(self,
                    container,
                    stream_path,
                    records,
                    access_key=None,
                    raise_for_status=None,
                    transport_actions=None):
        """Adds records to a stream.

        You can optionally assign a record to specific stream shard by specifying a related shard ID, or associate
        the record with a specific partition key to ensure that similar records are assigned to the same shard.
        By default, the platform assigns records to shards using a Round Robin algorithm. The max number of records
        is 1000.

        See https://www.iguazio.com/docs/latest-release/data-layer/reference/web-apis/streaming-web-api/putrecords/.

        Parameters
        ----------
        container (Required) : str
            The container on which to operate.
        stream_path (Required) : str
            The stream_path of the stream.
        records (Required) : []dict
            A list of dictionaries with the following keys:
            - shard_id: int, optional
                        The ID of the shard to which to assign the record, as an integer between 0 and one less than
                        the stream's shard count. When both ShardId and PartitionKey are set, the record is assigned
                        according to the shard ID, and PartitionKey is ignored. When neither a Shard ID or a partition
                        key is provided in the request, the platform's default shard-assignment algorithm is used.
            - data: str, required
                    Record data.
            - client_info: bytes/bytearray, optional
                           Custom opaque information that can optionally be provided by the producer.
                           This metadata can be used, for example, to save the data format of a record, or the time at
                           which a sensor or application event was triggered.
            - partition_key: str, optional
                             A partition key with which to associate the record (see Record Metadata). Records with the
                             same partition key are assigned to the same shard, subject to the following exceptions: if
                             a shard ID is also provided for the record (see the Records ShardId request parameter),
                             the record is assigned according to the shard ID, and PartitionKey is ignored. In addition,
                             if you increase a stream's shard count after its creation (see UpdateStream), new records
                             with a previously used partition key will be assigned either to the same shard that was
                             previously used for this partition key or to a new shard. All records with the same
                             partition key that are added to the stream after the shard-count change will be assigned
                             to the same shard (be it the previously used shard or a new shard). When neither a Shard
                             ID or a partition key is provided in the request, the platform's default shard-assignment
                             algorithm is used

            For example:
                [
                    {'shard_id': 1, 'data': 'first shard record #1'},
                    {'shard_id': 1, 'data': 'first shard record #2'},
                    {'shard_id': 10, 'data': 'invalid shard record #1'},
                    {'shard_id': 2, 'data': 'second shard record #1'},
                    {'data': 'some shard record #1'},
                ]

        Return Value
        ----------
        A `Response` object, whose `output` is `PutRecordsOutput`.
        """
        stream_path = self._ensure_path_ends_with_slash(stream_path)

        return self._transport.request(container,
                                       access_key or self._access_key,
                                       raise_for_status,
                                       transport_actions,
                                       v3io.dataplane.request.encode_put_records,
                                       locals(),
                                       v3io.dataplane.output.PutRecordsOutput)

    def get_records(self,
                    container,
                    stream_path,
                    shard_id,
                    location,
                    access_key=None,
                    raise_for_status=None,
                    transport_actions=None,
                    limit=None):
        """Retrieves (consumes) records from a stream shard.

        See https://www.iguazio.com/docs/latest-release/data-layer/reference/web-apis/streaming-web-api/getrecords/.

        Parameters
        ----------
        container (Required) : str
            The container on which to operate.
        stream_path (Required) : str
            The stream_path of the stream
        location (Required) : str
            The location within the shard at which to begin consuming records.
        access_key (Optional) : str
            The access key with which to authenticate. Defaults to the V3IO_ACCESS_KEY env.
        limit (Optional) : int
            The maximum number of records to return in the response. The minimum is 1. There's no restriction on
            the amount of returned records, but the maximum supported overall size of all the returned records is
            10 MB and the maximum size of a single record is 2 MB, so calculate the limit accordingly.

        Return Value
        ----------
        A `Response` object, whose `output` is `GetRecordsOutput`.
        """
        stream_path = self._ensure_path_ends_with_slash(os.path.join(stream_path, str(shard_id)))

        return self._transport.request(container,
                                       access_key or self._access_key,
                                       raise_for_status,
                                       transport_actions,
                                       v3io.dataplane.request.encode_get_records,
                                       locals(),
                                       v3io.dataplane.output.GetRecordsOutput)



# Copyright 2019 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import ujson
import xml.etree.ElementTree

import v3io.dataplane.transport


class HttpResponseError(Exception):
    """Exception raised on bad http status"""
    pass


class Response(object):

    def __init__(self, output, status_code, headers, body):
        self.status_code = status_code
        self.body = body
        self.headers = headers
        self._output = output
        self._parsed_output = None

    @property
    def output(self):
        if self._parsed_output:
            return self._parsed_output

        if self._output and self.body:
            try:
                # TODO: It's expensive to always try to parse as JSON first. Better use headers or a heuristic to decide the format.
                try:
                    parsed_output = ujson.loads(self.body)
                except Exception:
                    parsed_output = xml.etree.ElementTree.fromstring(self.body)
            except Exception:
                raise HttpResponseError(f"Failed to parse response with status {self.status_code}, "
                                        f"body {self.body}, headers={self.headers}")

            self._parsed_output = self._output(parsed_output)

            return self._parsed_output

    def raise_for_status(self, expected_statuses=None):
        if expected_statuses == v3io.dataplane.transport.RaiseForStatus.never:
            return

        # "always" and "none" are equivalent. use the one that's faster to compare against
        if expected_statuses == v3io.dataplane.transport.RaiseForStatus.always:
            expected_statuses = None

        if (expected_statuses is None and self.status_code >= 300) \
                or (expected_statuses and self.status_code not in expected_statuses):
            raise HttpResponseError('Request failed with status {0}: {1}'.format(self.status_code, self.body))


class Responses(object):

    def __init__(self):
        self.responses = []
        self.success = True

    def add_response(self, response):
        self.responses.append(response)

        if response.status_code != 200:
            self.success = False

    def raise_for_status(self):
        if not self.success:
            raise HttpResponseError('Failed to put items')



# Copyright 2019 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os

import v3io.dataplane.request
import v3io.dataplane.transport


class Transport(object):

    def __init__(self, logger, endpoint=None, max_connections=None, timeout=None, verbosity=None):
        self._logger = logger
        self._endpoint = self._get_endpoint(endpoint)
        self._timeout = timeout
        self.max_connections = max_connections or 8

        self._set_log_method(verbosity)

    def close(self):
        pass

    def requires_access_key(self):
        return False

    def restart(self):
        pass

    def request(self,
                container,
                access_key,
                raise_for_status,
                transport_actions,
                encoder,
                encoder_args,
                output=None):

        # default to sending/receiving
        transport_actions = transport_actions or v3io.dataplane.transport.Actions.send_and_receive

        # allocate a request
        request = v3io.dataplane.request.Request(container,
                                                 access_key,
                                                 raise_for_status,
                                                 encoder,
                                                 encoder_args,
                                                 output)

        # if all we had to do is encode, return now
        if transport_actions == v3io.dataplane.transport.Actions.encode_only:
            return request

        # send the request
        inflight_request = self.send_request(request)

        # wait for the response
        return self.wait_response(inflight_request)

    def send_request(self, request):
        return request

    @staticmethod
    def _get_endpoint(endpoint):

        if endpoint is None:
            endpoint = os.environ.get('V3IO_API')

            if endpoint is None:
                raise RuntimeError('Endpoints must be passed to context or specified in V3IO_API')

        if not endpoint.startswith('http://') and not endpoint.startswith('https://'):
            endpoint = 'http://' + endpoint

        return endpoint.rstrip('/')

    def _set_log_method(self, verbosity):
        # by default, the log method is null
        log_method = self._log_null

        if verbosity == 'DEBUG':
            log_method = self._log

        setattr(self, 'log', log_method)

    def _log(self, message, *args, **kw_args):
        self._logger.debug_with(message, *args, **kw_args)

    def _log_null(self, message, *args, **kw_args):
        pass



# Copyright 2019 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import aiohttp
import asyncio
import v3io.dataplane.request
import v3io.dataplane.response
import v3io.dataplane.transport


class Transport(object):

    def __init__(self, logger, endpoint=None, max_connections=None, timeout=None, verbosity=None, retry_intervals=None):
        self._logger = logger
        self._endpoint = self._get_endpoint(endpoint)
        self._timeout = timeout
        self.max_connections = max_connections or 8
        self._connector = aiohttp.TCPConnector()
        self._client_session = aiohttp.ClientSession(connector=self._connector)
        # spend ~1 min in retries before raising the exception to the user
        self.retry_intervals = retry_intervals or (0, 0, 0.1, 0.3, 1.0) + 12 * (5.0,)
        self._set_log_method(verbosity)

    async def close(self):
        await self._client_session.close()
        await self._connector.close()

    async def request(self,
                      container,
                      access_key,
                      raise_for_status,
                      encoder,
                      encoder_args,
                      output=None):

        # allocate a request
        request = v3io.dataplane.request.Request(container,
                                                 access_key,
                                                 raise_for_status,
                                                 encoder,
                                                 encoder_args,
                                                 output)

        path = request.encode_path()

        self.log('Tx', method=request.method, path=path, headers=request.headers, body=request.body)

        client_os_error_retry_counter = 0

        while (True):
            try:
                # call the encoder to get the response
                async with self._client_session.request(request.method,
                                                        self._endpoint + '/' + path,
                                                        headers=request.headers,
                                                        data=request.body,
                                                        ssl=False) as http_response:

                    # get contents
                    contents = await http_response.content.read()

                    # create a response
                    response = v3io.dataplane.response.Response(output,
                                                                http_response.status,
                                                                http_response.headers,
                                                                contents)

                    # enforce raise for status
                    response.raise_for_status(request.raise_for_status or raise_for_status)

                    self.log('Rx', status_code=response.status_code, headers=response.headers, body=contents)

                    return response
            except v3io.dataplane.response.HttpResponseError as response_error:
                self._logger.warn_with('Response error: {}'.format(str(response_error)))
                raise response_error
            except aiohttp.ClientOSError:
                client_os_error_retry_counter += 1
                if (client_os_error_retry_counter == len(self.retry_intervals)):
                    raise

            await asyncio.sleep(self.retry_intervals[client_os_error_retry_counter])

    @staticmethod
    def _get_endpoint(endpoint):

        if endpoint is None:
            endpoint = os.environ.get('V3IO_API')

            if endpoint is None:
                raise RuntimeError('Endpoints must be passed to context or specified in V3IO_API')

        if not endpoint.startswith('http://') and not endpoint.startswith('https://'):
            endpoint = 'http://' + endpoint

        return endpoint.rstrip('/')

    def _set_log_method(self, verbosity):
        # by default, the log method is null
        log_method = self._log_null

        if verbosity == 'DEBUG':
            log_method = self._log

        setattr(self, 'log', log_method)

    def _log(self, message, *args, **kw_args):
        self._logger.debug_with(message, *args, **kw_args)

    def _log_null(self, message, *args, **kw_args):
        pass



# Copyright 2019 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import sys
import ujson

import v3io.dataplane.transport.requests
import v3io.dataplane.transport.httpclient
import v3io.dataplane.request
import v3io.dataplane.batch
import v3io.dataplane.response
import v3io.dataplane.output
import v3io.dataplane.kv_cursor
import v3io.aio.dataplane.transport.aiohttp
import v3io.common.helpers
import v3io.logger


class Client(object):

    def __init__(self,
                 logger=None,
                 endpoint=None,
                 access_key=None,
                 max_connections=None,
                 timeout=None,
                 logger_verbosity=None,
                 transport_verbosity='info',
                 retry_intervals=None):
        """Creates a v3io client, used to access v3io

        Parameters
        ----------
        logger (Optional) : logger
            An optional pre-existing logger. If not passed, a logger is created with 'logger_verbosity' level
        endpoint (Optional) : str
            The v3io endpoint to connect to (e.g. http://v3io-webapi:8081). if empty, the env var
            V3IO_API is used
        access_key (Optional) : str
            The access key with which to authenticate. Defaults to the V3IO_ACCESS_KEY env. this can
            be overridden per request if needed
        max_connections (Optional) : int
            The number of connections to create towards v3io - defining the max number of parallel
            operations towards v3io. Defaults to 8
        timeout (Optional) : None
            For future use
        logger_verbosity (Optional) : INFO / DEBUG
            If 'logger' is not provided, this will specify the verbosity of the created logger.
        transport_verbosity (Optional) : INFO / DEBUG
            If set to 'DEBUG', transport will log lots of information at the cost of performance. It uses
            the "debug_with" logger interface, so wither a logger set to DEBUG level must be passed in 'logger' or
            'logger_verbosity' must be set to DEBUG
        retry_intervals (Optional) : tuple of float
            Tuple of intervals to use for exponential backoff in case of retries

        Return Value
        ----------
        A `Client` object
        """
        self._logger = logger or self._create_logger(logger_verbosity)
        self._access_key = access_key or os.environ.get('V3IO_ACCESS_KEY')

        if not self._access_key:
            raise ValueError('Access key must be provided in Client() arguments or in the '
                             'V3IO_ACCESS_KEY environment variable')

        self._transport = v3io.aio.dataplane.transport.aiohttp.Transport(self._logger,
                                                                         endpoint,
                                                                         max_connections,
                                                                         timeout,
                                                                         transport_verbosity,
                                                                         retry_intervals)

        # create models
        self.kv, self.object, self.stream, self.container = self._create_models()

    async def close(self):
        await self._transport.close()

    def _create_logger(self, logger_verbosity):
        logger = v3io.logger.Logger(level=logger_verbosity or 'INFO')
        logger.set_handler('stdout', sys.stdout, v3io.logger.HumanReadableFormatter())

        return logger

    @staticmethod
    def _get_schema_contents(key, fields):
        return ujson.dumps({
            'hashingBucketNum': 0,
            'key': key,
            'fields': fields
        })

    def _create_models(self):
        import v3io.aio.dataplane.object
        import v3io.aio.dataplane.kv
        import v3io.aio.dataplane.stream
        import v3io.aio.dataplane.container

        # create models
        return v3io.aio.dataplane.kv.Model(self), \
               v3io.aio.dataplane.object.Model(self), \
               v3io.aio.dataplane.stream.Model(self), \
               v3io.aio.dataplane.container.Model(self)
